/********************************************************************/
/********************************************************************/

/* fl_ext_color_chooser()
 *
 * This code released into the public domain by Don Zimmer
 * with permission from Baird Software, October 2004
 * provided as-is, use at your own risk
 *
 */

/********************************************************************/
/********************************************************************/


#include "fl_ext_color_chooser.h"


#ifdef WIN32
  #include <FL/Fl.H>
  #include <FL/x.H>
  #include <commdlg.h>    /* includes common dialog functionality */
  #include <dlgs.h>       /* includes common dialog template defines */
  #include <cderr.h>      /* includes the common dialog error codes */
  #include "win32_templates.h"
  #include <ctype.h>
  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h>
#else
  #include <FL/Fl.H>
  #include <FL/Fl_Color_Chooser.H>
#endif


/********************************************************************/
/********************************************************************/


#ifdef WIN32

typedef struct {
  char *pTitle;
  BYTE alpha;
  int bAlpha;
} CS_SPEC;

#endif


/********************************************************************/
/********************************************************************/


#ifdef WIN32

  static UINT CALLBACK true_color_hook(HWND hdlg, UINT uiMsg, WPARAM wParam, LPARAM lParam);

  static COLORREF alpha_blend(COLORREF crBase, COLORREF crOverlay, BYTE alpha);

  static void draw_true_color_box(HWND hdlg, WPARAM wParam, LPARAM lParam, BYTE alpha);

  static void init_dialog(HWND hdlg, char *pTitle, int bAlpha, BYTE alpha);

  static void invalidate_current_color(HWND hdlg);

  static void set_input(HWND hWnd, int nVal);

  static void handle_alpha_input(HWND hdlg, BYTE *pAlpha);

#endif


/********************************************************************/
/* externals                                                        */
/********************************************************************/


int fl_ext_color_chooser(unsigned int *pcolor, char *pTitle /* = NULL */, int bAlpha /* = 0 */)
{
  int rc;
  unsigned char *pix = (unsigned char*)pcolor;

#ifndef WIN32
  
  rc = fl_color_chooser(pTitle, pix[0], pix[1], pix[2]);

#else

  CHOOSECOLOR spec;
  CS_SPEC cs_spec;

  static COLORREF cust_col[NUM_CUSTOM_COLORS] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };

  cs_spec.pTitle = pTitle;
  cs_spec.alpha = pix[3];
  cs_spec.bAlpha = bAlpha;

  spec.lStructSize = sizeof(CHOOSECOLOR);
  spec.hwndOwner = GetForegroundWindow();
  spec.hInstance = (HWND)GetModuleHandle(NULL);
  spec.rgbResult = *pcolor;

  spec.lpCustColors = cust_col;

  spec.Flags  = CC_RGBINIT|CC_FULLOPEN|CC_ANYCOLOR|CC_SOLIDCOLOR;
  
  spec.Flags |= CC_ENABLEHOOK;
  
  if(bAlpha) 
  {
    spec.Flags |= CC_ENABLETEMPLATE;
    spec.lpTemplateName  = MAKEINTRESOURCE(IDD_TPL_TRUECOLOR);
  }

  spec.lCustData       = (LPARAM)&cs_spec;
  spec.lpfnHook        = true_color_hook;

  spec.rgbResult &= 0X00FFFFFF; 

  rc = ::ChooseColor(&spec);

  if(rc) 
  {
    *pcolor = spec.rgbResult;
    pix[3] = cs_spec.alpha;
  }

#endif

  return rc;
}


/********************************************************************/
/********************************************************************/
/********************************************************************/
/********************************************************************/
#ifdef WIN32
/********************************************************************/
/********************************************************************/
/********************************************************************/
/********************************************************************/


/********************************************************************/
/* handler                                                          */
/********************************************************************/


static UINT CALLBACK true_color_hook(HWND hdlg, UINT uiMsg, WPARAM wParam, LPARAM lParam)
{
  static BYTE def_alpha = 255, *pAlpha = &def_alpha;
  static int bAlpha = 0;
  UINT rc = 0;

  switch (uiMsg)
  {
  case WM_INITDIALOG:
    {
      CHOOSECOLOR *pSpec = (CHOOSECOLOR*)lParam;
      CS_SPEC *pCsSpec = (CS_SPEC*)pSpec->lCustData;
      pAlpha = &pCsSpec->alpha;
      bAlpha = pCsSpec->bAlpha;

      init_dialog(hdlg, pCsSpec->pTitle, bAlpha, pCsSpec->alpha);

      rc = TRUE; /* so initial focus is set by the default dialog procedure */
    }
    break;

  case WM_CTLCOLORSTATIC:
    if(bAlpha)
    {
      if (::GetDlgItem(hdlg, COLOR_CURRENT) == (HWND)lParam)
      {
        draw_true_color_box(hdlg, wParam, lParam, *pAlpha);
      }
    }
    break;

  case WM_CTLCOLOREDIT:
    if(bAlpha)
    {
      if (::GetDlgItem(hdlg, COLOR_ALPHA) == (HWND)lParam)
      {
        handle_alpha_input(hdlg, pAlpha);
        invalidate_current_color(hdlg);
      }
    }
    break;

  case WM_COMMAND:
    if(bAlpha)
    {
      if (EN_KILLFOCUS == HIWORD(wParam) && COLOR_ALPHA == LOWORD(wParam))
      {
        /* in case field is blank
        */
        set_input((HWND)lParam, *pAlpha);
      }
    }
  }

  /* almost always return zero, we want the default handlers to kick-in also
  */
  return rc;
}


/********************************************************************/
/* support                                                          */
/********************************************************************/


static COLORREF alpha_blend(COLORREF crBase, COLORREF crOverlay, BYTE alpha)
{
  COLORREF cr;
  BYTE *dst      = (BYTE*)&cr;
  BYTE *src1     = (BYTE*)&crBase;
  BYTE *src2     = (BYTE*)&crOverlay;
  double a_frac1 = (255.0 - alpha)/255.0;
  double a_frac2 = alpha/255.0;

  dst[0] = (BYTE)((src1[0]*a_frac1) + (src2[0]*a_frac2) + 0.5);
  dst[1] = (BYTE)((src1[1]*a_frac1) + (src2[1]*a_frac2) + 0.5);
  dst[2] = (BYTE)((src1[2]*a_frac1) + (src2[2]*a_frac2) + 0.5);
  dst[3] = 0X00;

  return cr;
}


/********************************************************************/
/********************************************************************/


static void draw_true_color_box(HWND hdlg, WPARAM wParam, LPARAM lParam, BYTE alpha)
{
  HDC       hDC = (HDC)wParam;
  RECT      rect;
  COLORREF  cr;
  HBRUSH    hBrush;
  HGDIOBJ   hOldBrush;
  int       r, g, b;
  char      s[42];

  ::GetWindowText(::GetDlgItem(hdlg, COLOR_RED), s, 42);
  r = atoi(s);

  ::GetWindowText(::GetDlgItem(hdlg, COLOR_GREEN), s, 42);
  g = atoi(s);

  ::GetWindowText(::GetDlgItem(hdlg, COLOR_BLUE), s, 42);
  b = atoi(s);

  COLORREF crCurrent     = 0X00000000;
  ((BYTE*)&crCurrent)[0] = (BYTE)r;
  ((BYTE*)&crCurrent)[1] = (BYTE)g;
  ((BYTE*)&crCurrent)[2] = (BYTE)b;

  rect.left = 19;
  rect.top = 0;
  rect.right = 39;
  rect.bottom = 25;
  ::MapDialogRect(hdlg, &rect);

  cr = alpha_blend(0X00FFFFFF, crCurrent, alpha);
  hBrush = ::CreateSolidBrush(cr);
  hOldBrush = ::SelectObject(hDC, hBrush);
  ::FillRect(hDC, &rect, hBrush);
  ::DeleteObject(hBrush);

  rect.left   += 5;
  rect.top    += 5;
  rect.right  -= 6;
  rect.bottom -= 6;

  cr = alpha_blend(0X00000000, crCurrent, alpha);
  hBrush = ::CreateSolidBrush(cr);
  ::FrameRect(hDC, &rect, hBrush);
  ::DeleteObject(hBrush);

  ::SelectObject(hDC, hOldBrush);
}


/********************************************************************/
/********************************************************************/


static void init_dialog(HWND hdlg, char *pTitle, int bAlpha, BYTE alpha)
{
  if(pTitle) ::SetWindowText(hdlg, pTitle);

  ::ShowWindow(::GetDlgItem(hdlg, COLOR_MIX), SW_HIDE);

  if(bAlpha)
  {
    char s[42];
    sprintf(s, "%i", alpha);
    ::SetWindowText(::GetDlgItem(hdlg, COLOR_ALPHA), s);
  }
}


/********************************************************************/
/********************************************************************/


/* used this function because SendDlgItemMessage and SendMessage 
 * had problems with the size of the DC rectangle
 */
static void invalidate_current_color(HWND hdlg)
{
  char s[42];

  HWND redInput = ::GetDlgItem(hdlg, COLOR_RED);

  ::GetWindowText(redInput, s, 42);
  ::SetWindowText(redInput, s);   
}


/********************************************************************/
/********************************************************************/


static void set_input(HWND hWnd, int nVal)
{
  char s[42];

  sprintf(s, "%i", nVal);

  ::SetWindowText(hWnd, s);   
}


/********************************************************************/
/********************************************************************/


/* only empty string and 0 to 255 allowed
*/
static void handle_alpha_input(HWND hdlg, BYTE *pAlpha)
{
  char s[42];
  int  l, i, a;

  HWND alphaInput = ::GetDlgItem(hdlg, COLOR_ALPHA);

  ::GetWindowText(alphaInput, s, 42);
  
  /* blank is ok, though
  */
  if (!s[0]) return;

  /* invalid characters will reset value and select all
  */
  l = strlen(s);
  for (i = 0; i < l; i++)
  {
    if(!isdigit(s[i]))
    {
      set_input(alphaInput, *pAlpha);
      ::SendDlgItemMessage(hdlg, COLOR_ALPHA, EM_SETSEL, 0, -1);
      return;
    }
  }

  /* range check
  */
  a = atoi(s);
  if (a > 255) 
  {
    a = 255;
    set_input(alphaInput, a);
  }

  *pAlpha = (BYTE)a;
}


/********************************************************************/
/********************************************************************/
/********************************************************************/
/********************************************************************/
#endif /* WIN32 */
/********************************************************************/
/********************************************************************/
/********************************************************************/
/********************************************************************/



