/********************************************************************/
/********************************************************************/

/* fl_ext_date_chooser()
 *
 * This code released into the public domain by Don Zimmer
 * with permission from Baird Software, October 2004
 * provided as-is, use at your own risk
 *
 */

/********************************************************************/
/********************************************************************/


#include "fl_ext_date_chooser_ui.h"
#include "fl_ext_date_chooser.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

#include "mjd.h"


#ifndef WIN32
#define _timezone __timezone
#endif


#define SECONDS_2_DAYS(t) ((t)/(24.0*60.0*60.0))


/********************************************************************/
/********************************************************************/


static const char *english_headings[7] = { "S", "M", "T", "W", "T", "F", "S" };
static const char *english_prompt = "Select Date";
static const char *english_ok = "OK";
static const char *english_cancel = "Cancel";


/********************************************************************/
/********************************************************************/


static const char *dd_am_pm_24[3] = {
  "- AM", 
  "- PM", 
  "- 24"
};

enum { FMT_AM, FMT_PM, FMT_24 };


/********************************************************************/
/********************************************************************/


static const char **dd_months   = mjd_long_month;
static const char **dd_headings = english_headings;
static const char  *dd_prompt   = english_prompt;
static const char  *dd_ok       = english_ok;
static const char  *dd_cancel   = english_cancel;

static char dd_days[42][3];

static int dd_fmt = FMT_24;
static int dd_fmt_lock = 0;
static int dd_fmt_mod = 3;

static int dd_rc = -1;

static int dd_year, dd_month, dd_day, dd_hour, dd_minute, dd_second;

static Fl_Ext_DateChooser_UI *dd_pDlg = NULL;


/********************************************************************/
/********************************************************************/


static void change_format(int val);
static void change_hour(int val);
static void refresh(void);


/********************************************************************/
/* externals                                                        */
/********************************************************************/


/* set language-specific form text - use NULL to reset to english
*/
void fl_ext_date_text(
  const char *prompt, 
  const char *months[12], 
  const char *headings[7], 
  const char *ok, 
  const char *cancel)
{
  dd_prompt   = (prompt)?   prompt   : english_prompt;
  dd_months   = (months)?   months   : mjd_long_month; // english_months;
  dd_headings = (headings)? headings : english_headings;
  dd_ok       = (ok)?       ok       : english_ok;
  dd_cancel   = (cancel)?   cancel   : english_cancel;
}


/********************************************************************/
/********************************************************************/


/* set whether to limit to am/pm or 24-hour formats
*/
void fl_ext_date_format(int am_pm, int lock)
{
  dd_fmt = (am_pm)? FMT_AM : FMT_24;
  dd_fmt_lock = lock;
  dd_fmt_mod = (dd_fmt_lock)? ((am_pm)? 2 : 1) : 3;
}


/********************************************************************/
/********************************************************************/


/* select a date, send DD_MIN to initialize with current date/time
*/
int fl_ext_date_chooser(double *date)
{
  Fl_Ext_DateChooser_UI dlg;
  Fl_Widget **pDays;
  Fl_Widget **pHeadings;
  double date_in, dSecond;
  int i;

  if(!date) return 0;

  date_in = *date;

  if(date_in <= DD_MIN)
  {
    time_t now = time(NULL);

    localtime(&now); // hack:  initializes _timezone variable

    date_in = MJD_1970 + SECONDS_2_DAYS(now - _timezone);
  } 

  mjd_calc_gregorian(MJD_ROUND(date_in, 0), &dd_year, &dd_month, &dd_day, &dd_hour, &dd_minute, &dSecond);

  dd_second = (int)dSecond;

  if(FMT_AM == dd_fmt)
  {
    if(dd_hour >= 12) dd_fmt = FMT_PM;
  }
  else if(FMT_PM == dd_fmt)
  {
    if(dd_hour < 12) dd_fmt = FMT_AM;
  }

  /* setup form
  */
  dd_pDlg = &dlg;
  pDays = (Fl_Widget **)dd_pDlg->m_grpDays->array();
  pHeadings = (Fl_Widget **)dd_pDlg->m_grpHeadings->array();

  for(i=0; i<42; i++) pDays[i]->label(dd_days[i]);

  for(i=0; i<12; i++) dd_pDlg->m_drpMonth->add(dd_months[i]);

  for(i=0; i<7; i++) pHeadings[i]->label(dd_headings[i]);

  dd_pDlg->m_btnTimeUp->clear_visible_focus();
  dd_pDlg->m_btnTimeDown->clear_visible_focus();
  dd_pDlg->m_btnYearUp->clear_visible_focus();
  dd_pDlg->m_btnYearDown->clear_visible_focus();

  if(dd_fmt_lock && dd_fmt == FMT_24)
  {
    dd_pDlg->m_inpFormat->deactivate();
  }
  else
  {
    dd_pDlg->m_inpFormat->activate();
  }

  dd_pDlg->m_btnOK->label(dd_ok);
  dd_pDlg->m_btnCancel->label(dd_cancel);
  dd_pDlg->m_wndMain->label(dd_prompt);
  dd_pDlg->m_wndMain->set_modal();

  refresh();

  dd_pDlg->m_wndMain->show();

  dd_rc = -1;

  while(dd_rc < 0) Fl::check(); /* Note: make sure that ESC is disabled */

  dd_pDlg->m_wndMain->hide();

  dd_pDlg = NULL;

  if(dd_rc == 1) *date = mjd_calc_julian(dd_year, dd_month, dd_day, dd_hour, dd_minute, dd_second);

  return (dd_rc == 1);
}


/********************************************************************/
/* callbacks for FLTK dialog                                        */
/********************************************************************/


void dd_done(int rc)
{
  dd_rc = rc;
}


/********************************************************************/
/********************************************************************/


void dd_change_year(int val)
{
  if(val > 0 && val < 2100)
  {
    dd_year = val;
    dd_day = 1;
  }
  refresh();
}


/********************************************************************/
/********************************************************************/


void dd_spin_year(int val)
{
  dd_pDlg->m_inpYear->take_focus();
  dd_change_year(dd_year + val);
}


/********************************************************************/
/********************************************************************/


void dd_change_month(int val)
{
  if(val > 0 && val < 13)
  {
    dd_month = val;
    dd_day = 1;
  }
  refresh();
}


/********************************************************************/
/********************************************************************/


void dd_change_day(int val)
{
  int bLeap, nDays;
  double dDate;

  dDate = mjd_calc_julian(dd_year, dd_month, dd_day, dd_hour, dd_minute, dd_second);

  bLeap = MJD_LEAPYEAR(dd_year);

  nDays = mjd_days_in_month[bLeap][dd_month-1];

  if(val > 0 && val <= nDays)
  {
    dd_day = val;
  }

  refresh();

  /* fix for fltk loss of focus oddity (related to radio buttons in a group)
  */
  {
    Fl_Widget **pDays = (Fl_Widget **)dd_pDlg->m_grpDays->array();

    for(int i=0; i<42; i++)
    {
      if(atoi(pDays[i]->label()) == dd_day)
      {
        pDays[i]->take_focus();
        break;
      }
    }
  }
}


/********************************************************************/
/********************************************************************/


void dd_change_hour(int val)
{
  if(val < 0 || val >= 24)
  {
    refresh();
  }
  else
  {
    switch(dd_fmt)
    {
    case FMT_AM:
      if(12 == val) val = 0;
      else if(val > 12) dd_fmt = FMT_24;
      break;

    case FMT_PM:
      if(0 == val || val > 12) dd_fmt = FMT_24;
      else if(val < 12) val += 12;
      break;
    }

    change_hour(val);
  }
}


/********************************************************************/
/********************************************************************/


void dd_change_minute(int val)
{
  if(val >=0 && val < 60)
  {
    dd_minute = val;
  }
  refresh();
}


/********************************************************************/
/********************************************************************/


void dd_change_second(int val)
{
  if(val >=0 && val < 60)
  {
    dd_second = val;
  }
  refresh();
}


/********************************************************************/
/********************************************************************/


void dd_change_format(const char *val)
{
  int i = 0;

  while(val[i])
  {
    if('a' == val[i] || 'A' == val[i])
    {
      change_format(FMT_AM);
      return;
    }
    else if('p' == val[i] || 'P' == val[i])
    {
      change_format(FMT_PM);
      return;
    }
    else if('2' == val[i])
    {
      change_format(FMT_24);
      return;
    }

    i++;
  }
}


/********************************************************************/
/********************************************************************/


void dd_spin_time(int val)
{
  Fl_Widget *pFocus = Fl::focus();

  if(pFocus == dd_pDlg->m_inpFormat)
  {
    change_format((dd_fmt + val + dd_fmt_mod) % dd_fmt_mod);
  }
  else if(pFocus == dd_pDlg->m_inpSecond)
  {
    dd_change_second((dd_second + val + 60) % 60);
  }
  else if(pFocus == dd_pDlg->m_inpMinute)
  {
    dd_change_minute((dd_minute + val + 60) % 60);
  }
  else // use dd_pDlg.m_inpHour
  {
    dd_pDlg->m_inpHour->take_focus();

    int new_hour = (dd_hour + val + 24) % 24;

    switch(dd_fmt)
    {
    case FMT_AM:
      new_hour = new_hour % 12;
      break;

    case FMT_PM:
      new_hour = 12 + (new_hour % 12);
      break;
    }

    change_hour(new_hour);
  }
}


/********************************************************************/
/* support                                                          */
/********************************************************************/


static void change_hour(int val)
{
  if(val >=0 && val < 24)
  {
    dd_hour = val;
  }
  refresh();
}


/********************************************************************/
/********************************************************************/


static void change_format(int val)
{
  /* when locked, can't change to or from FMT_24
  */
  int invalid = (dd_fmt_lock && ((FMT_24 == dd_fmt) || (FMT_24 == val)));

  if(!invalid)
  {
    dd_fmt = val;

    switch(dd_fmt)
    {
    case FMT_AM:
      dd_hour = dd_hour % 12;
      break;

    case FMT_PM:
      dd_hour = 12 + (dd_hour % 12);
      break;
    }
  }

  refresh();
}


/********************************************************************/
/********************************************************************/


static void refresh(void)
{
  Fl_Widget **pDays;
  Fl_Button *pDay;
  char s[42];
  int i, bLeap, nDays, iFirst;
  double dDate, dFirst;

  /* calc date params
  */
  dDate = mjd_calc_julian(dd_year, dd_month, dd_day, dd_hour, dd_minute, dd_second);

  bLeap = MJD_LEAPYEAR(dd_year);

  nDays = mjd_days_in_month[bLeap][dd_month-1];

  dFirst = mjd_calc_julian(dd_year, dd_month, 1, 0, 0, 0);

  iFirst = MJD_WEEKDAY(dFirst);

  /* refresh form
  */
  pDays = (Fl_Widget **)dd_pDlg->m_grpDays->array();
  pDay = (Fl_Button*)(pDays[iFirst + dd_day - 1]);

  /* dd_year 
  */
  sprintf(s, "%04d", dd_year);
  dd_pDlg->m_inpYear->value(s);

  /* dd_month
  */
  dd_pDlg->m_drpMonth->value(dd_month - 1);

  /* dd_day
  */
  for(i=0; i<42; i++)
  {
    sprintf(dd_days[i], "");
    pDays[i]->deactivate();
  }
  for(i=0; i<nDays; i++)
  {
    sprintf(dd_days[iFirst + i], "%d", i+1);
    pDays[iFirst + i]->activate();
  }
  pDay->setonly();

  /* dd_hour
  */
  sprintf(s, "%02d", (dd_fmt != FMT_24)? MJD_12HR(dd_hour) : dd_hour);
  dd_pDlg->m_inpHour->value(s);

  /* dd_minute
  */
  sprintf(s, "%02d", dd_minute);
  dd_pDlg->m_inpMinute->value(s);

  /* dd_second
  */
  sprintf(s, "%02d", dd_second);
  dd_pDlg->m_inpSecond->value(s);

  /* am/pm/24
  */
  dd_pDlg->m_inpFormat->value(dd_am_pm_24[dd_fmt]);

  #ifdef DEBUG
  printf("%s, %s %02d, %04d %02d:%02d:%02d\n", 
    mjd_long_weekday[MJD_WEEKDAY(dDate)], 
    mjd_long_month[dd_month - 1],
    dd_day,
    dd_year, 
    dd_hour,
    dd_minute,
    dd_second);
  #endif
}


/********************************************************************/
/********************************************************************/





