//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// Fl_Ext_TreeBrowser
//
// This code released into the public domain by Don Zimmer
// with permission from Baird Software, October 2004
// provided as-is, use at your own risk

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


#include <stdio.h>

#include <FL/fl_draw.H>
#include <FL/Fl_Window.H>

#include "fl_ext_tree_browser.h"


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


#define ITEM(p) ((Fl_Ext_TreeItem*)p)


//////////////////////////////////////////////////////////////////////
// construction
//////////////////////////////////////////////////////////////////////


Fl_Ext_TreeBrowser::Fl_Ext_TreeBrowser(int x, int y, int w, int h, const char *l /* =0 */)
                   :Fl_Browser_(x, y, w, h, l)
{
  type(FL_HOLD_BROWSER);

  when(FL_WHEN_RELEASE);
  Fl_Widget::callback(_s_DefaultCallback);

  m_pFirst = m_pLast = NULL;
}


Fl_Ext_TreeBrowser::~Fl_Ext_TreeBrowser()
{
  _Remove(NULL, 1);
}


//////////////////////////////////////////////////////////////////////
// fltk overrides
//////////////////////////////////////////////////////////////////////


void *Fl_Ext_TreeBrowser::item_first() const 
{
  return m_pFirst;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


void *Fl_Ext_TreeBrowser::item_next(void *l) const 
{
  return ITEM(l)->m_pNext;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


void *Fl_Ext_TreeBrowser::item_prev(void *l) const 
{
  return ITEM(l)->m_pPrev;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


int Fl_Ext_TreeBrowser::item_width(void *v) const 
{
  Fl_Ext_TreeItem *pItem = ITEM(v);

  if (!pItem->m_bVisible) return 0;

  int w = 3 + pItem->m_nLevel*10;

  if (pItem->IsParent()) 
  {
    w += 14; // extra 4 for space after box
  }
  else if (pItem->m_nLevel)
  {
    w += 4; // extra 4 to line up with parent
  }

  int bGrey = 0;
  if(pItem->GetGrey) bGrey = pItem->GetGrey(pItem->m_pData);
  if(bGrey)
  {
    fl_font(textfont()+FL_ITALIC, textsize());
  }
  else
  {
    fl_font(textfont(), textsize());
  }

  if(pItem->GetName)
  {
    w += int(fl_width(pItem->GetName(pItem->m_pData)));
  }
  else
  {
    w += int(fl_width(" (empty)"));
  }

  return w;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


int Fl_Ext_TreeBrowser::item_height(void *l) const 
{
  return (ITEM(l)->m_bVisible)?textsize()+2:0; 
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// retain open/closed array
//
void Fl_Ext_TreeBrowser::item_draw(void *v, int x, int y, int, int) const 
{
  Fl_Ext_TreeItem *pItem = ITEM(v);

  // left edge
  //
  x += 3 + pItem->m_nLevel * 10;

  // text x
  int tx = x;

  // parent x
  int px = x - 10;

  // colour
  //
  if(pItem->m_bSelected)
  {
    fl_color(fl_contrast(FL_BLACK,FL_SELECTION_COLOR));
  }
  else 
  {
    fl_color(FL_BLACK);
  }

  // tree branch lines
  //
  if(pItem->IsParent()) 
  {
    // box with '-' inside
    fl_rect(x, y+3, 9, 9);
    fl_line(x+2, y+7, x+6, y+7);

    if (pItem->IsOpen()) 
    {
      // south connector
      fl_line(x+4, y+11, x+4, y+13); 
    } 
    else 
    {
      // turn '-' into '+'
      fl_line(x+4, y+5, x+4, y+9); 
    }

    if(pItem->m_nLevel)
    {
      // west connector to parent
      fl_line(px+4, y+7, x, y+7);
    }

    tx += 14; // extra 4 for space after box
  }
  else if(pItem->m_nLevel) 
  {
    tx += 4; // extra 4 to line up with parent

    // west connector to parent
    fl_line(px+4, y+7, px+8, y+7);
  }

  // north and south connetors to parent
  //
  if(pItem->m_nLevel)
  {
    if(pItem->IsLastSibling())
    {
      fl_line(px+4, y, px+4, y+7); 
    }
    else
    {
      fl_line(px+4, y, px+4, y+13); 
    }
  }

  // disconnected brach lines from previous levels, start up two
  //
  Fl_Ext_TreeItem *pParent = pItem->GetParent();
  while(pParent)
  {
    px -= 10;
    if(pParent->m_nLevel)
    {
      if(!pParent->IsLastSibling())
      {
        fl_line(px+4, y, px+4, y+13);
      }
    }
    pParent = pParent->GetParent();
  }
  
  // draw text
  //
  int bGrey = 0;
  if(pItem->GetGrey) bGrey = pItem->GetGrey(pItem->m_pData);
  if(bGrey)
  {
    if(pItem->m_bSelected)
    {
      fl_color(fl_contrast(FL_INACTIVE_COLOR,FL_SELECTION_COLOR));
    }
    else 
    {
      fl_color(FL_INACTIVE_COLOR);
    }

    fl_font(textfont()+FL_ITALIC, textsize());
  }
  else
  {
    if(pItem->m_bSelected)
    {
      fl_color(fl_contrast(textcolor(),FL_SELECTION_COLOR));
    }
    else 
    {
      fl_color(textcolor());
    }

    fl_font(textfont(), textsize());
  }

  if(pItem->GetName)
  {
    fl_draw(pItem->GetName(pItem->m_pData), tx, y+13);
  }
  else
  {
    fl_draw("(empty)", tx, y+13);
  }
}


/////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////


int Fl_Ext_TreeBrowser::handle(int event) 
{
  Fl_Ext_TreeItem *pItem;
  int x, y, w, h, button;

  switch (event) 
  {
  case FL_PUSH:
    bbox(x,y,w,h);
    if(!Fl::event_inside(x,y,w,h)) break;
    
    x = Fl::event_x();
    y = Fl::event_y();
    button = Fl::event_button();

    pItem = ITEM(find_item(y));
    if(!pItem) return 0;

    if(!Select(pItem)) return 0;

    // hide/show
    if(pItem->IsParent() && (x > (pItem->m_nLevel*10+3)) && (x <= (pItem->m_nLevel*10+13)) )
    {
      pItem->SetBranchVisible(!pItem->IsOpen());
      redraw();
    }

    if(Fl::event_clicks())
    {
      if(pItem->HandleInvoke) pItem->HandleInvoke(pItem->m_pData);
    }
    else if(button == FL_RIGHT_MOUSE)
    {
      if(pItem->HandleMenu) 
      {
        Fl_Browser_::handle(event);
        pItem->HandleMenu(pItem->m_pData); // could redraw based upon rc
        return 0;
      }
    }
    break;

  case FL_DRAG: // disable FL_DRAG
    return 1;

  case FL_SHORTCUT: // disable arrow keys until ready to tackle them
  case FL_KEYBOARD:
    switch (Fl::event_key()) 
    {
    case FL_Enter:
    case FL_Down:
    case FL_Up:
	    return 1;
    }
    break;

  default:
    break;
  }

  return Fl_Browser_::handle(event);
}


//////////////////////////////////////////////////////////////////////
// additions
//////////////////////////////////////////////////////////////////////


int Fl_Ext_TreeBrowser::Add(Fl_Ext_TreeItem *pItem, int bSelect /* =0 */)
{
  Fl_Ext_TreeItem *pNext;
  int rc;

  if(!m_pFirst) 
  {
    m_pFirst = pItem;

    m_pLast = pItem;
    pNext = pItem->m_pNext;
    while(pNext) { m_pLast = pNext; pNext = pNext->m_pNext; }
  }
  else 
  {
    pItem->m_pPrev = m_pLast;
    m_pLast->m_pNext = pItem;

    m_pLast = pItem;
    pNext = pItem->m_pNext;
    while(pNext) { m_pLast = pNext; pNext = pNext->m_pNext; }
  }

  // show siblings
  Fl_Ext_TreeItem *pParent = pItem->GetParent();
  if(pParent && pItem->m_bVisible) pParent->SetBranchVisible(1);

  rc = 1;
  if(bSelect) rc = Select(pItem);

  redraw();
  return rc;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


int Fl_Ext_TreeBrowser::Remove(Fl_Ext_TreeItem *pItem) 
{ 
  return _Remove(pItem, 1); 
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


int Fl_Ext_TreeBrowser::Prune(Fl_Ext_TreeItem *pItem) 
{ 
  return _Remove(pItem, 0); 
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


int Fl_Ext_TreeBrowser::InsertAfter(Fl_Ext_TreeItem *pPrev, Fl_Ext_TreeItem *pItem, int bSelect /* =0 */)
{
  if(!pItem) return 0;
  if(!pPrev) return Add(pItem, bSelect);

  Fl_Ext_TreeItem *pHead = pPrev;
  Fl_Ext_TreeItem *pTail = pPrev->m_pNext;

  pHead->m_pNext = pItem;
  pItem->m_pPrev = pHead;

  Fl_Ext_TreeItem *pNext = pItem;
  while(pNext->m_pNext)
  {
    pNext = pNext->m_pNext;
  }

  pNext->m_pNext = pTail;
  if(pTail) pTail->m_pPrev = pNext;
  else m_pLast = pNext;

  // show siblings
  Fl_Ext_TreeItem *pParent = pItem->GetParent();
  if(pParent && pItem->m_bVisible) pParent->SetBranchVisible(1);

  int rc = 1;
  if(bSelect) rc = Select(pItem);

  redraw();
  return rc;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


// remove pTreeItem and all children, freeing memory
//
int Fl_Ext_TreeBrowser::_Remove(Fl_Ext_TreeItem *pItem, int bFree)
{
  Fl_Ext_TreeItem *pNext;

  Fl_Ext_TreeItem *pSelected = ITEM(selection());

  if(!pItem) // entire tree
  {
    pItem = m_pFirst;
    if(!pItem) return 0;

    Select(ITEM(NULL));
    m_pFirst = m_pLast = NULL;
    new_list();

    if(bFree)
    {
      while(pItem)
      {
        pNext = pItem->m_pNext;
        delete pItem;
        pItem = pNext;
      }
    }
  }
  else // just this branch
  {
    Fl_Ext_TreeItem *pPrev = pItem->m_pPrev;

    int nLevel = pItem->m_nLevel;

    // cap off pruned branch at head
    if(!bFree) pItem->m_pPrev = NULL;

    while(pItem)
    {
      pNext = pItem->m_pNext;

      if(pItem == pSelected) Select(ITEM(NULL));

      if(bFree) delete pItem; // deletes data also

      pItem = pNext;

      if(pItem)
      {
        if(pItem->m_nLevel <= nLevel)
        {
          // cap off pruned branch at tail
          if(!bFree) pItem->m_pPrev->m_pNext = NULL;
          
          // end of branch - done
          break;
        }
      }
    }


    // fix up links
    if(pItem)
    {
      pItem->m_pPrev = pPrev;

      if(pPrev) 
      {
        pPrev->m_pNext = pItem;
      }
      else
      {
        m_pFirst = pItem; 
      }
    }
    else
    {
      if(pPrev) 
      {
        pPrev->m_pNext = NULL;
        m_pLast = pPrev;
      }
      else
      {
        m_pFirst = m_pLast = NULL;
      }
    }

    // fix up selection
    if(!selection() && pSelected)
    {
      if(pPrev)
      {
        if(pPrev->m_pNext) Select(pPrev->m_pNext);
        else Select(pPrev);
      }
      else if(m_pFirst) 
      {
        Select(m_pFirst);
      }
    }

    redraw();
  }

  return 1;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


int Fl_Ext_TreeBrowser::Select(Fl_Ext_TreeItem *pItem)
{
  Fl_Ext_TreeItem *pSelected = ITEM(selection());

  if(pSelected == pItem) return 1;

  // deselect old
  //
  if(pSelected)
  {
    if(pSelected->HandleSelect) 
    {
      if(!pSelected->HandleSelect(pSelected->m_pData, 0)) 
      {
        if(pItem) return 0; // allow Select(NULL) to pass
      }
    }
    pSelected->m_bSelected = 0;
  }

  // select new
  //
  if(pItem)
  {
    if(pItem->HandleSelect)
    {
      if(!pItem->HandleSelect(pItem->m_pData, 1))
      {
        if(pSelected) // re-select old
        {
          if(pSelected->HandleSelect) pSelected->HandleSelect(pSelected->m_pData, 1);
          pSelected->m_bSelected = 1;
        }
        return 0;
      }
    }

    // make sure we can see the selection 
    //
    Fl_Ext_TreeItem *pParent = pItem->GetParent();
    while(pParent) { pParent->SetBranchVisible(1); pParent = pParent->GetParent(); }
    
    pItem->m_bSelected = 1;
    select(pItem);
  }
  else
  {
    deselect();
  }

  redraw();
  return 1;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


// selects the nth visible line
//
int Fl_Ext_TreeBrowser::Select(int nLine)
{
  Fl_Ext_TreeItem *pItem = m_pFirst;

  int i = 0;

  while(pItem)
  {
    if(pItem->m_bVisible) i++;

    if(i == nLine) break;

    pItem = pItem->m_pNext;
  }

  if(pItem) return Select(pItem);
  return 0;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


// selects the last visible line at a level
//
int Fl_Ext_TreeBrowser::SelectLast(int nLevel)
{
  Fl_Ext_TreeItem *pItem = m_pLast;

  while(pItem)
  {
    if(pItem->m_bVisible && pItem->m_nLevel == nLevel) break;

    pItem = pItem->m_pPrev;
  }

  if(pItem) return Select(pItem);
  return 0;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


Fl_Ext_TreeItem *Fl_Ext_TreeBrowser::Selection(void) const
{
  return (Fl_Ext_TreeItem*)selection();
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


Fl_Ext_TreeItem *Fl_Ext_TreeBrowser::Find(int (*pFunc)(void *a, void *b), void *pArg)
{
  if(!pFunc) pFunc = _s_DefaultFindFunc;

  Fl_Ext_TreeItem *pItem = m_pFirst;

  while(pItem)
  {
    if(pFunc(pItem->m_pData, pArg)) break;

    pItem = pItem->m_pNext;
  }

  return pItem;
}


//////////////////////////////////////////////////////////////////////
// support
//////////////////////////////////////////////////////////////////////


int Fl_Ext_TreeBrowser::_s_DefaultFindFunc(void *a, void *b)
{
  return (a == b)?1:0;
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


// default do-nothing callback, user template
//
void Fl_Ext_TreeBrowser::_s_DefaultCallback(Fl_Widget *o, void *)
{
#ifdef DEBUG
  Fl_Ext_TreeBrowser *pBrowser = (Fl_Ext_TreeBrowser *)o;

  Fl_Ext_TreeItem *pItem = ITEM(pBrowser->selection());
  if(!pItem) return;

  Fl_Widget *pParent = pBrowser;
  while(pParent->parent()) pParent = pParent->parent();
  void *pUI = ((Fl_Window*)pParent)->user_data();

  printf("UI %i Brs: %i Item %i Lev %i Parent %i Open %i Data %i\n", 
    pUI, 
    pBrowser, 
    pItem, 
    pItem->m_nLevel, 
    pItem->IsParent(), 
    pItem->IsOpen(),
    pItem->m_pData);
#endif
}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////




