/***********************************************************************/
/***********************************************************************/

/* winblit()
 *
 * Send an image of a window to a printer using a configured PRINT_SPEC
 *
 * This code released into the public domain by Don Zimmer
 * October 2004
 * provided as-is, use at your own risk
 *
 */

/***********************************************************************/
/***********************************************************************/


#include <windows.h>
#include "fl_ext_page_setup.h" /* for PRINT_SPEC */


/***********************************************************************/
/***********************************************************************/


int winblit(PRINT_SPEC *sp, HWND hwnd, int border)
{
  RECT rect;
  int win_w, win_h;
  HDC win_dc, print_dc;
  OSVERSIONINFO osvi;
  char *driver;
  LPDEVMODE devmode;
  double lm, tm, print_w, print_h, win_aspect, print_aspect;
  int px, py, pw, ph;
  DOCINFO info;
  int rc = 0;
  
  /* printer setup
  */
  devmode = (LPDEVMODE)GlobalLock(sp->hDevMode);
  if(!devmode) return 0;
  osvi.dwOSVersionInfoSize = sizeof(osvi);
  GetVersionEx(&osvi);
  driver = (osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)? NULL : "WINSPOOL";
  print_dc = CreateDC(driver, sp->queue, NULL, devmode);
  GlobalUnlock(sp->hDevMode);
  if(!print_dc) return 0;

  /* print position (centered)
  */
  if(border) GetWindowRect(hwnd, &rect);
  else GetClientRect(hwnd, &rect);
  win_w = rect.right - rect.left;
  win_h = rect.bottom - rect.top;
  win_aspect = win_w/(double)win_h;
  lm = sp->margin[PAGE_LM];
  tm = sp->margin[PAGE_TM];
  print_w = sp->size[PAGE_W] - lm - sp->margin[PAGE_RM];
  print_h = sp->size[PAGE_H] - tm - sp->margin[PAGE_BM];
  print_aspect = print_w/print_h;
  if(win_aspect > print_aspect)
  {
    double hh = print_w / win_aspect;
    tm += (print_h - hh) / 2;
    print_h = hh;
  }
  else
  {
    double ww = print_h * win_aspect;
    lm += (print_w - ww) / 2;
    print_w = ww;
  }
  if(PAGE_CM == sp->eUnits)
  {
    print_w = CM_TO_INCH(print_w);
    print_h = CM_TO_INCH(print_h);
    lm = CM_TO_INCH(lm);
    tm = CM_TO_INCH(tm);
  }
  px = (int)(lm * sp->dpi[PAGE_W]) - GetDeviceCaps(print_dc, PHYSICALOFFSETX);
  py = (int)(tm * sp->dpi[PAGE_H]) - GetDeviceCaps(print_dc, PHYSICALOFFSETY);
  pw = (int)(print_w * sp->dpi[PAGE_W]);
  ph = (int)(print_h * sp->dpi[PAGE_H]);

  /* do it
  */
  win_dc = (border)? GetWindowDC(hwnd) : GetDC(hwnd);
  if(win_dc)
  {
    HDC hdc_bmp = 0;
    HBITMAP hbitmap;
    BITMAPINFOHEADER bmih;
    void *pbits;

    /* grab the window into a DIB, no error checking
    */
    hdc_bmp = CreateCompatibleDC(win_dc);
    bmih.biSize          = sizeof(BITMAPINFOHEADER);
    bmih.biWidth         = win_w;
    bmih.biHeight        = win_h;
    bmih.biPlanes        = 1;
    bmih.biBitCount      = 24;
    bmih.biCompression   = BI_RGB;
    bmih.biSizeImage     = 0;
    bmih.biXPelsPerMeter = 0; 
    bmih.biYPelsPerMeter = 0; 
    bmih.biClrUsed       = 0;
    bmih.biClrImportant  = 0;
    hbitmap = CreateDIBSection(hdc_bmp, (BITMAPINFO*)&bmih, DIB_RGB_COLORS, &pbits, NULL, 0);
    hbitmap = (HBITMAP)SelectObject(hdc_bmp, hbitmap);
    BitBlt(hdc_bmp, 0, 0, win_w, win_h, win_dc, 0,0, SRCCOPY);

    /* print bitmap
    */
    info.cbSize = sizeof(DOCINFO);
    info.lpszDocName = "winblit";
    info.lpszOutput = NULL;
    info.lpszDatatype = NULL;
    info.fwType = 0L;
    if(StartDoc(print_dc, &info))
    {
      if(StartPage(print_dc))
      {
        if(SetStretchBltMode(print_dc, COLORONCOLOR))
        {
          if(StretchBlt(print_dc, px, py, pw, ph, hdc_bmp, 0, 0, win_w, win_h, SRCCOPY))
          {
            rc = 1;
          }
        }
        EndPage(print_dc);
      }
      EndDoc(print_dc);
    }

    SelectObject(hdc_bmp, NULL);
    DeleteObject(hbitmap);
    DeleteDC(hdc_bmp);

    ReleaseDC(hwnd, win_dc);
  }

  DeleteDC(print_dc);

  return rc;
}


/***********************************************************************/
/***********************************************************************/


